#include "config.h"

#include "panelfront.h"
#include "displaynetworksort.h"

#if (defined(HAVE_LIBNCURSES) && defined(HAVE_LIBPANEL) && defined(BUILD_PANEL))

char *KismetHelpText[] = {
    "KISMET PANELS INTERFACE",
    "  Key  Action",
    "   z   Toggle fullscreen zoom of network view",
    "   m   Toggle muting of sound and speech",
    "   t   Tag (or untag) selected network",
    "   g   Group tagged networks",
    "   u   Ungroup current group",
    "",
    "POPUP WINDOWS",
    "   h   Help (What you're looking at now)",
    "   n   Name current network",
    "   i   Detailed information about selected network",
    "   s   Sort network list",
    "   l   Show wireless card power levels",
    "   d   Dump printable strings",
    "   x   Close popup window",
    "",
    "   q   Quit",
    NULL
};
#define HELP_SIZE 20

// Narrow text
char *KismetHelpTextNarrow[] = {
    "KISMET PANELS INTERFACE",
    "Key Action",
    " z  Toggle fullscreen net list",
    " m  Toggle muting",
    " t  Tag (or untag) selected",
    " g  Group tagged networks",
    " u  Ungroup current group",
    "",
    "POPUP WINDOWS",
    "   h   Help",
    "   n   Name network",
    "   i   Detailed information",
    "   s   Sort network list",
    "   l   Show signal levels",
    "   d   Dump printable strings",
    "   x   Close popup window",
    "",
    "   q   Quit",
    NULL
};
#define HELP_SIZE_NARROW 20

char *KismetSortText[] = {
    "Key  Sort                Key  Sort",
    " a   Auto-fit (standard)  c   Channel",
    " f   First time seen      F   First time seen (descending)",
    " l   Latest time seen     L   Latest time seen (descending)",
    " b   BSSID                B   BSSID (descending)",
    " s   SSID                 S   SSID (descending)",
    " p   Packet count         P   Packet count (descending)",
    " w   WEP                  x   Cancel",
    NULL
};
#define SORT_SIZE 10

char *KismetSortTextNarrow[] = {
    "Key Sort        Key Sort",
    " a  Auto-fit     c  Channel",
    " f  First time   F  First time (d)",
    " l  Latest time  L  Latest time (d)",
    " b  BSSID        B  BSSID (d)",
    " s  SSID         S  SSID (d)",
    " p  Packet count P  Packet count (d)",
    " w  WEP          x  Cancel",
    NULL
};
#define SORT_SIZE_NARROW 10

PanelFront::PanelFront() {
    errstr[0] = '\0';

    client = NULL;

    sub_sort = 0;
    sub_dump = 0;
    clear_dump = 0;

    hsize = COLS;
    vsize = LINES;

    cutoff = 0;

    muted = 0;
}

int PanelFront::InitDisplay(int in_decay, time_t in_start) {
    start_time = in_start;

    decay = in_decay;

    initscr();

    netwin = newwin(LINES-statheight, COLS-infowidth, 0, 0);

    infowin = newwin(LINES-statheight, infowidth, 0, 0);

    statuswin = newwin(statheight, COLS, 0, 0);
    scrollok(statuswin, 1);

    network_panel = new_panel(netwin);
    info_panel = new_panel(infowin);
    status_panel = new_panel(statuswin);

    noecho();
    cbreak();
    nodelay(netwin, true);
    nodelay(infowin, true);
    nodelay(statuswin, true);
    keypad(netwin, true);
    keypad(infowin, true);
    keypad(statuswin, true);

    move_panel(network_panel, 0, 0);
    move_panel(info_panel, 0, COLS-infowidth);
    move_panel(status_panel, LINES-statheight, 0);

    top_panel(network_panel);
    cur_panel = network_panel;

    zoomed = 0;

    muted = 0;

    return 0;
}

void PanelFront::RescaleDisplay() {
    wresize(netwin, LINES-statheight, COLS-infowidth);
    replace_panel(network_panel, netwin);

    wresize(infowin, LINES-statheight, infowidth);
    replace_panel(info_panel, infowin);

    wresize(statuswin, statheight, COLS);
    replace_panel(status_panel, statuswin);

    move_panel(network_panel, 0, 0);
    move_panel(info_panel, 0, COLS-infowidth);
    move_panel(status_panel, LINES-statheight, 0);
}

void PanelFront::SetColumns(string in_columns) {
    unsigned int begin = 0;
    unsigned int end = in_columns.find(",");

    column_vec.erase(column_vec.begin(), column_vec.end());

    while (end < in_columns.size()) {
        string opt = in_columns.substr(begin, end-begin);
        begin = end+1;
        end = in_columns.find(",", begin);

        column_vec.push_back(opt);
    }

    column_vec.push_back(in_columns.substr(begin, in_columns.size()));
}

int PanelFront::WriteStatus(string status) {
    wmove(statuswin, 1, 0);
    winsertln(statuswin);
    mvwaddstr(statuswin, 1, 2, status.substr(0, COLS-4).c_str());
}

int PanelFront::DrawDisplay() {
    if (hsize != COLS || vsize != LINES) {
        hsize = COLS; vsize = LINES;
        RescaleDisplay();
    }

    // The status window takes care of itself
    box(statuswin, '|', '-');
    mvwaddstr(statuswin, 0, 2, "Status");

    werase(netwin);
    box(netwin, '|', '-');

    mvwaddstr(netwin, 0, 2, "Networks");

    werase(infowin);
    box(infowin, '|', '-');
    mvwaddstr(infowin, 0, 2, "Info");

    mvwaddch(infowin, LINES-statheight-1, 3, 'H');
    mvwaddch(infowin, LINES-statheight-1, 5, 'M');
    mvwaddch(infowin, LINES-statheight-1, 7, 'S');

    // Now draw the info window
    char info[infowidth-2];

    mvwaddstr(infowin, 2, 2, "Ntwrks");
    snprintf(info, infowidth-2, "%6d", client->FetchNumNetworks());
    mvwaddstr(infowin, 3, 2, info);

    mvwaddstr(infowin, 4, 2, "Pckets");
    snprintf(info, infowidth-2, "%6d", client->FetchNumPackets());
    mvwaddstr(infowin, 5, 2, info);

    mvwaddstr(infowin, 6, 2, "Cryptd");
    snprintf(info, infowidth-2, "%6d", client->FetchNumCrypt());
    mvwaddstr(infowin, 7, 2, info);

    mvwaddstr(infowin, 8, 2, "  Weak");
    snprintf(info, infowidth-2, "%6d", client->FetchNumInteresting());
    mvwaddstr(infowin, 9, 2, info);

    mvwaddstr(infowin, 10, 2, " Noise");
    snprintf(info, infowidth-2, "%6d", client->FetchNumNoise());
    mvwaddstr(infowin, 11, 2, info);

    mvwaddstr(infowin, 12, 2, "Discrd");
    snprintf(info, infowidth-2, "%6d", client->FetchNumDropped());
    mvwaddstr(infowin, 13, 2, info);

    mvwaddstr(infowin, LINES-statheight-2, 2, "Elapsd");
    time_t elapsed = client->FetchTime() - start_time;
    snprintf(info, infowidth-2, "%02d%02d%02d",
             (int) (elapsed / 60) / 60, (int) (elapsed / 60) % 60,
             (int) elapsed % 60);
    mvwaddstr(infowin, infowin->_maxy, 2, info);

    // Max networks displayable on the panel as it is now
    max_display = netwin->_maxy - 3;
    print_width = netwin->_maxx - 4;

    int drop;

    // One:  Get our new data from the client
    PopulateGroups();
    // Two:  Recalculate all our agregate data
    UpdateGroups();
    // Three: Copy it to our own local vector so we can sort it.
    vector<display_network *> display_vector = group_vec;

    //vector<wireless_network *> network_vector = client->FetchNetworkList();

    char sortxt[24];
    sortxt[0] = '\0';

    switch (sortby) {
    case sort_auto:
        // Trim it ourselves for autofit mode.
        // This is really easy because we don't allow groups to be expanded in autofit
        // mode, so we just make it fit.

        snprintf(sortxt, 24, "(Autofit)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortLastTimeLT());

        drop = display_vector.size() - max_display - 1;

        if (drop > 0) {
            display_vector.erase(display_vector.begin(), display_vector.begin() + drop);
        }
        sort(display_vector.begin(), display_vector.end(), DisplaySortFirstTimeLT());
        display_start = 0;

        break;

        // Otherwise, go on to sort the groups by the other means available to us...
    case sort_channel:
        snprintf(sortxt, 24, "(Channel)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortChannel());
        break;
    case sort_first:
        snprintf(sortxt, 24, "(First Seen)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortFirstTimeLT());
        break;
    case sort_first_dec:
        snprintf(sortxt, 24, "(First Seen desc)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortFirstTime());
        break;
    case sort_last:
        snprintf(sortxt, 24, "(Latest Seen)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortLastTimeLT());
        break;
    case sort_last_dec:
        snprintf(sortxt, 24, "(Latest Seen desc)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortLastTime());
        break;
    case sort_bssid:
        snprintf(sortxt, 24, "(BSSID)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortBSSIDLT());
        break;
    case sort_bssid_dec:
        snprintf(sortxt, 24, "(BSSID desc)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortBSSID());
        break;
    case sort_ssid:
        snprintf(sortxt, 24, "(SSID)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortSSIDLT());
        break;
    case sort_ssid_dec:
        snprintf(sortxt, 24, "(SSID desc)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortSSID());
        break;
    case sort_wep:
        snprintf(sortxt, 24, "(WEP)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortWEP());
        break;
    case sort_packets:
        snprintf(sortxt, 24, "(Packets)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortPacketsLT());
        break;
    case sort_packets_dec:
        snprintf(sortxt, 24, "(Packets desc)");

        sort(display_vector.begin(), display_vector.end(), DisplaySortPackets());
        break;
    }

    last_displayed.erase(last_displayed.begin(), last_displayed.begin() + last_displayed.size());
//    last_displayed = display_vector;
//    last_displayed = network_vector;

    int num = 0;
    int voffset = 2;

    // Print the headers
    int pos = 4;

    for (unsigned int col = 0; col < column_vec.size(); col++) {
        char title[1024];
        int len = 0;

        if (column_vec[col] == "decay") {
            snprintf(title, 1024, " ");
            len = 1;
        } else if (column_vec[col] == "name") {
            snprintf(title, 1024, "Name");
            len = 25;
        } else if (column_vec[col] == "shortname") {
            snprintf(title, 1024, "Name");
            len = 15;
        } else if (column_vec[col] == "ssid") {
            snprintf(title, 1024, "SSID");
            len = 25;
        } else if (column_vec[col] == "shortssid") {
            snprintf(title, 1024, "SSID");
            len = 15;
        } else if (column_vec[col] == "type") {
            snprintf(title, 1024, "T");
            len = 1;
        } else if (column_vec[col] == "wep") {
            snprintf(title, 1024, "W");
            len = 1;
        } else if (column_vec[col] == "channel") {
            snprintf(title, 1024, "Ch");
            len = 2;
        } else if (column_vec[col] == "data") {
            snprintf(title, 1024, " Data");
            len = 5;
        } else if (column_vec[col] == "llc") {
            snprintf(title, 1024, "  LLC");
            len = 5;
        } else if (column_vec[col] == "crypt") {
            snprintf(title, 1024, "Crypt");
            len = 5;
        } else if (column_vec[col] == "weak") {
            snprintf(title, 1024, "Weak");
            len = 5;
        } else if (column_vec[col] == "bssid") {
            snprintf(title, 1024, "BSSID");
            len = 17;
        } else if (column_vec[col] == "flags") {
            snprintf(title, 1024, "Flags");
            len = 5;
        } else if (column_vec[col] == "ip") {
            snprintf(title, 1024, "IP Range");
            len = 15;
        } else if (column_vec[col] == "mask") {
            snprintf(title, 1024, "IP Mask");
            len = 15;
        } else if (column_vec[col] == "gateway") {
            snprintf(title, 1024, "IP Gateway");
            len = 15;
        } else if (column_vec[col] == "packets") {
            snprintf(title, 1024, "Packts");
            len = 6;
        } else if (column_vec[col] == "info") {
            snprintf(title, 1024, "Beacon Info");
            len = 15;
        } else if (column_vec[col] == "maxrate") {
            snprintf(title, 1024, "Rate");
            len = 5;
        }

        if (pos + len > print_width+1)
            break;

        mvwaddstr(netwin, 1, pos, title);

        pos += len + 1;

    }

    cutoff = 0;

    for (unsigned int i = display_start; i < display_vector.size(); i++) {

        if (display_vector[i]->networks.size() == 0)
            continue;

    //    if (!paused)
        last_displayed.push_back(display_vector[i]);

        wireless_network *net = &display_vector[i]->virtnet;

        if (i == (display_start + selected) && sortby != sort_auto) {
            wattron(netwin, A_REVERSE);
            char bar[1024];
            memset(bar, ' ', 1024);
            int w = print_width;
            if (w >= 1024)
                w = 1024;
            bar[w] = '\0';
            mvwaddstr(netwin, num+voffset, 2, bar);
        }

        string netline;

        // Build the netline for the group or single host and tag it for expansion if
        // appropriate for this sort and group
        netline = NetLine(net,
                          display_vector[i]->name == "" ? display_vector[i]->virtnet.ssid.c_str() : display_vector[i]->name.c_str(),
                          0,
                          display_vector[i]->type == group_host ? 0 : 1,
                          sortby == sort_auto ? 0 : display_vector[i]->expanded,
                          display_vector[i]->tagged);

        mvwaddstr(netwin, num+voffset, 1, netline.c_str());

        if (i == (display_start + selected) && sortby != sort_auto)
            wattroff(netwin, A_REVERSE);

        num++;
        display_end = i;

        if (num > max_display)
            break;

        if (sortby == sort_auto || display_vector[i]->type != group_bundle ||
            display_vector[i]->expanded == 0)
            continue;

        // Now we're not hilighted, and we've made sure we'll fit past the end of the screen,
        // so if we're not autofitting and we're expanded we want to
        for (unsigned int y = 0; y < display_vector[i]->networks.size(); y++) {
            net = display_vector[i]->networks[y];

            netline = NetLine(net, net->ssid.c_str(), 1, 0, 0, 0);

            mvwaddstr(netwin, num+voffset, 1, netline.c_str());

            num++;

            if (num > max_display) {
                cutoff = 1;
                break;
            }
        }

        if (num > max_display)
            break;
    }

    if ((display_start + selected) > display_end) {
        selected = display_end - display_start;
        DrawDisplay();
    }


    /*
    if (paused) {
        wattron(netwin, A_BLINK);
        mvwaddstr(netwin, 0, 12, " PAUSED ");
        wattroff(netwin, A_BLINK);
    } else {
        mvwaddstr(netwin, 0, 12, sortxt);
        }
        */

    mvwaddstr(netwin, 0, 12, sortxt);

    if (display_start != 0 && sortby != sort_auto) {
        mvwaddstr(netwin, 0, netwin->_maxx - 10, "(-) Up");
    }

    if (display_end < group_vec.size()-1 && sortby != sort_auto) {
        mvwaddstr(netwin, netwin->_maxy,
                  netwin->_maxx - 10, "(+) Down");
    }

    if (cur_panel == popup_panel) {
        werase(panel_window(popup_panel));
        (this->*printer)(popup_panel);

        // This is kind of a nasty trick to play, but if the popup handled user
        // info then it's probably made itself go away by now
        if (popup_panel != NULL) {
            box(panel_window(popup_panel), '|', '-');
            mvwaddstr(panel_window(popup_panel), 0, 2, popup_title.c_str());

            if (popup_paused) {
                wattron(panel_window(popup_panel), A_BLINK);
                mvwaddstr(panel_window(popup_panel), 0, popup_title.size() + 4, " PAUSED ");
                wattroff(panel_window(popup_panel), A_BLINK);
            }

            if (popup_start != 0) {
                mvwaddstr(panel_window(popup_panel), 0, panel_window(popup_panel)->_maxx - 10,
                          "(-) Up");
            }

            if (popup_end < popup_length-1 && popup_end != 0) {
                mvwaddstr(panel_window(popup_panel), panel_window(popup_panel)->_maxy,
                          panel_window(popup_panel)->_maxx - 10,
                          "(+) Down");
            }
        }
    }

    char gpsdata[1024];
//    if (gps != NULL) {
    float lat, lon, alt, spd;
    int mode;

    client->FetchLoc(&lat, &lon, &alt, &spd, &mode);

    if (!(lat == 0 && lon == 0 && alt == 0 && spd == 0 && mode == 0)) {

        char fix[16];

        if (mode == -1)
            snprintf(fix, 16, "No signal");
        else if (mode == 2)
            snprintf(fix, 5, "2D");
        else if (mode == 3)
            snprintf(fix, 5, "3D");
        else
            snprintf(fix, 5, "NONE");

        snprintf(gpsdata, 1024, "Lat %.3f Lon %.3f Alt %.1f Spd %.3f Fix %s",
                 lat, lon, alt, spd, fix);

        mvwaddstr(netwin, netwin->_maxy, 2, gpsdata);

        // fprintf(stderr, "found: %f %f %f %f\n", lat, lon, alt, spd);

    }

    wmove(panel_window(cur_panel), 0, 0);

    update_panels();
    doupdate();

    return 1;
}

// This argument list is getting increasingly nasty, hopefully I'll be able to
// come back and clean all this up
string PanelFront::NetLine(wireless_network *net, const char *name, int sub, int group, int expanded, int tagged) {
    char retchr[4096];
    char tmpchr[4096];

    memset(retchr, 0, 4096);
    memset(tmpchr, 0, 4096);

    // Sanity check just to be sure
    if (print_width > 4096)
        print_width = 4096;

    if (tagged)
        snprintf(retchr, 4096, "*");
    else if (group && expanded)
        snprintf(retchr, 4096, "-");
    else if (group && !expanded)
        snprintf(retchr, 4096, "+");
    else if (sub)
        snprintf(retchr, 4096, "|");

    int pos = 2;
    for (unsigned int col = 0; col < column_vec.size(); col++) {
        char element[1024];
        int len = 0;

        if (column_vec[col] == "decay") {
            if ((client->FetchTime() - net->last_time) < decay)
                snprintf(element, 1024, "!");
            else if ((client->FetchTime() - net->last_time) < (decay * 2))
                snprintf(element, 1024, ".");
            else
                snprintf(element, 1024, " ");
            len = 1;
        } else if (column_vec[col] == "name") {
            if (net->cloaked) {
                snprintf(element, 26, "<%s>", name);
            } else {
                snprintf(element, 26, "%s", name);
            }
            len = 25;
        } else if (column_vec[col] == "shortname") {
            if (net->cloaked) {
                snprintf(element, 16, "<%s>", name);
            } else {
                snprintf(element, 16, "%s", name);
            }
            len = 15;
        } else if (column_vec[col] == "ssid") {
            if (net->cloaked) {
                snprintf(element, 26, "<%s>", net->ssid.c_str());
            } else {
                snprintf(element, 26, "%s", net->ssid.c_str());
            }
            len = 25;
        } else if (column_vec[col] == "shortssid") {
            if (net->cloaked) {
                snprintf(element, 16, "<%s>", net->ssid.c_str());
            } else {
                snprintf(element, 16, "%s", net->ssid.c_str());
            }
            len = 15;
        } else if (column_vec[col] == "type") {
            if (group)
                snprintf(element, 1024, "G");
            else if (net->type == network_ap)
                snprintf(element, 1024, "A");
            else if (net->type == network_adhoc)
                snprintf(element, 1024, "H");
            else if (net->type == network_probe)
                snprintf(element, 1024, "P");
            else if (net->type == network_data)
                snprintf(element, 1024, "D");
            else
                snprintf(element, 1024, "?");

            len = 1;
        } else if (column_vec[col] == "wep") {
            if (net->wep)
                snprintf(element, 1024, "Y");
            else
                snprintf(element, 1024, "N");
            len = 1;
        } else if (column_vec[col] == "channel") {
            snprintf(element, 3, "%02d", net->channel);
            len = 2;
        } else if (column_vec[col] == "data") {
            snprintf(element, 6, "%5d", net->data_packets);
            len = 5;
        } else if (column_vec[col] == "llc") {
            snprintf(element, 6, "%5d", net->llc_packets);
            len = 5;
        } else if (column_vec[col] == "crypt") {
            snprintf(element, 6, "%5d", net->crypt_packets);
            len = 5;
        } else if (column_vec[col] == "weak") {
            snprintf(element, 6, "%5d", net->interesting_packets);
            len = 5;
        } else if (column_vec[col] == "packets") {
            snprintf(element, 7, "%6d", net->data_packets + net->llc_packets + net->crypt_packets);
            len = 6;
        } else if (column_vec[col] == "bssid") {
            snprintf(element, 18, "%s", net->bssid.c_str());
            len = 17;
        } else if (column_vec[col] == "info") {
            snprintf(element, 16, "%s", net->beacon_info.c_str());
            len = 15;
        } else if (column_vec[col] == "flags") {
            char atype;
            if (net->atype == address_dhcp)
                atype = 'D';
            else if (net->atype == address_arp)
                atype = 'A';
            else if (net->atype == address_udp)
                atype = 'U';
            else if (net->atype == address_group)
                atype = 'G';
            else
                atype = ' ';

            snprintf(element, 6, "%c%c%c%c%c",
                     atype,
                     (net->octets != 0 && net->octets != 4) ? net->octets + '0' : ' ',
                     net->cisco_equip.size() > 0 ? 'C' : ' ',
                     ' ',
                     ' ');
            len = 5;
        } else if (column_vec[col] == "ip") {
            if (net->atype == address_none) {
                snprintf(element, 1024, "0.0.0.0");
            } else {
                snprintf(element, 16, "%d.%d.%d.%d",
                         net->range_ip[0], net->range_ip[1],
                         net->range_ip[2], net->range_ip[3]);
            }
            len = 15;
        } else if (column_vec[col] == "mask") {
            if (net->atype == address_none) {
                snprintf(element, 1024, "0.0.0.0");
            } else {
                snprintf(element, 16, "%d.%d.%d.%d",
                         net->mask[0], net->mask[1],
                         net->mask[2], net->mask[3]);
            }
            len = 15;
        } else if (column_vec[col] == "gateway") {
            if (net->atype == address_none) {
                snprintf(element, 1024, "0.0.0.0");
            } else {
                snprintf(element, 16, "%d.%d.%d.%d",
                         net->gate_ip[0], net->gate_ip[1],
                         net->gate_ip[2], net->gate_ip[3]);
            }
            len = 15;
        } else if (column_vec[col] == "maxrate") {
            snprintf(element, 6, "%2.1f", net->maxrate);
            len = 5;
        }

        if (pos + len > print_width)
            break;

//        fprintf(stderr, "%s ... %s\n", retchr, element);

        snprintf(tmpchr, 4096, "%*s %s", (-1) * pos, retchr, element);

        strncpy(retchr, tmpchr, 4096);

//        snprintf(retchr, 4096, "%*s %s", (-1) * pos, retchr, element);

//        mvwaddstr(netwin, num+voffset, pos, element);

        pos += len + 1;
    }

    string ret = retchr;

    return ret;
}

int PanelFront::EndDisplay() {
    endwin();


    return 1;
}

void PanelFront::ZoomNetworks() {
    if (zoomed == 0) {
        wresize(netwin, LINES, COLS);
        replace_panel(network_panel, netwin);
        zoomed = 1;
    } else {
        wresize(netwin, LINES-statheight, COLS-infowidth);
        replace_panel(network_panel, netwin);

        if (selected > (unsigned int) netwin->_maxy - 3)
            selected = netwin->_maxy - 3;

        zoomed = 0;
    }
    DrawDisplay();
}

void PanelFront::SpawnPopup(string in_title, panel_printer in_print, int in_x = -1, int in_y = -1) {
    if (cur_panel == network_panel) {

        if (in_x == -1)
            in_x = LINES-5;
        if (in_y == -1)
            in_y = COLS-8;

        popupwin = newwin(in_x, in_y, 0, 0);

        popup_panel = new_panel(popupwin);
        nodelay(popupwin, true);
        keypad(popupwin, true);

        move_panel(popup_panel, (LINES/2) - (in_x/2), (COLS/2) - (in_y/2));

        printer = in_print;

        show_panel(popup_panel);

        popup_title = in_title;

        popup_start = popup_end = popup_length = 0;

        cur_panel = popup_panel;
        DrawDisplay();
    }

}

void PanelFront::PrintKismetHelp(PANEL *in_panel) {
    int x = 0;
    while (1) {
        if (KismetHelpText[x] == NULL)
            break;
        if (panel_window(in_panel)->_maxx < 54)
            mvwaddstr(panel_window(in_panel), 1+x, 2, KismetHelpTextNarrow[x]);
        else
            mvwaddstr(panel_window(in_panel), 1+x, 2, KismetHelpText[x]);
        x++;
    }
}

void PanelFront::PrintKismetSort(PANEL *in_panel) {
    int x = 0;
    while (1) {
        if (KismetSortText[x] == NULL)
            break;
        if (panel_window(in_panel)->_maxx < 64)
            mvwaddstr(panel_window(in_panel), 1+x, 2, KismetSortTextNarrow[x]);
        else
            mvwaddstr(panel_window(in_panel), 1+x, 2, KismetSortText[x]);
        x++;
    }

}

void PanelFront::PrintKismetPower(PANEL *in_panel) {
    int quality, power, noise;

    quality = client->FetchQuality();
    power = client->FetchPower();
    noise = client->FetchNoise();

    if (quality == -1 && power == -1 && noise == -1) {
        mvwaddstr(panel_window(in_panel), 2, 2, "Server did not report card power levels.");
        mvwaddstr(panel_window(in_panel), 3, 2, "No card information is available.");
        return;
    }

    int width = panel_window(in_panel)->_maxx - 10;

    if (width <= 5) {
        return;
    }

    char *bar = new char[width+1];

    if (quality > LINKQ_MAX)
        quality = LINKQ_MAX;
    if (power > LEVEL_MAX)
        power = LEVEL_MAX;
    if (noise > NOISE_MAX)
        noise = NOISE_MAX;

    double qperc = 0, pperc = 0, nperc = 0;
    if (quality != 0)
        qperc = (double) quality/LINKQ_MAX;
    if (power != 0)
        pperc = (double) power/LEVEL_MAX;
    if (noise != 0)
        nperc = (double) noise/NOISE_MAX;

    int qbar = 0, pbar = 0, nbar = 0;
    qbar = (int) (width * qperc);
    pbar = (int) (width * pperc);
    nbar = (int) (width * nperc);

    memset(bar, '=', width);
    memset(bar, 'X', qbar);
    bar[width] = '\0';
    mvwaddstr(panel_window(in_panel), 1, 2, "Q:");
    mvwaddstr(panel_window(in_panel), 1, 5, bar);

    memset(bar, '=', width);
    memset(bar, 'X', pbar);
    bar[width] = '\0';
    mvwaddstr(panel_window(in_panel), 2, 2, "P:");
    mvwaddstr(panel_window(in_panel), 2, 5, bar);

    memset(bar, '=', width);
    memset(bar, 'X', nbar);
    bar[width] = '\0';
    mvwaddstr(panel_window(in_panel), 3, 2, "N:");
    mvwaddstr(panel_window(in_panel), 3, 5, bar);

    snprintf(bar, width, "%d", quality);
    mvwaddstr(panel_window(in_panel), 1, width+6, bar);
    snprintf(bar, width, "%d", power);
    mvwaddstr(panel_window(in_panel), 2, width+6, bar);
    snprintf(bar, width, "%d", noise);
    mvwaddstr(panel_window(in_panel), 3, width+6, bar);

}

void PanelFront::PrintKismetInfo(PANEL *in_panel) {
    vector<string> details_text;
    char output[1024];

    werase(panel_window(in_panel));

    const int print_width = panel_window(in_panel)->_maxx - 5;

    if (details_network->name == "")
        snprintf(output, print_width, "Name    : %s", details_network->virtnet.ssid.c_str());
    else
        snprintf(output, print_width, "Name    : %s", details_network->name.c_str());
    details_text.push_back(output);

    for (unsigned int x = 0; x < details_network->networks.size(); x++) {
        wireless_network *dnet = details_network->networks[x];

        if (details_network->networks.size() > 1) {
            if (x != 0) {
                snprintf(output, print_width, " ");
                details_text.push_back(output);
            }
            snprintf(output, print_width, "Network %d", x+1);
            details_text.push_back(output);
        }

        // Convert the current details network into a vector of strings, so that
        // we can scroll it
        snprintf(output, print_width, "SSID    : %s", dnet->ssid.c_str());
        details_text.push_back(output);
    
        if (dnet->cloaked) {
            snprintf(output, print_width, "          SSID Cloaking on/Closed Network");
            details_text.push_back(output);
        }
    
        snprintf(output, print_width, "BSSID   : %s", dnet->bssid.c_str());
        details_text.push_back(output);
    
        snprintf(output, print_width, "Max Rate: %2.1f", dnet->maxrate);
        details_text.push_back(output);
    
        snprintf(output, print_width, "First   : %s", ctime((const time_t *) &dnet->first_time));
        details_text.push_back(output);
    
        snprintf(output, print_width, "Latest  : %s", ctime((const time_t *) &dnet->last_time));
        details_text.push_back(output);
    
        switch (dnet->type) {
        case network_ap:
            snprintf(output, print_width, "Type    : Access Point (infrastructure)");
            break;
        case network_adhoc:
            snprintf(output, print_width, "Type    : Ad-hoc");
            break;
        case network_probe:
            snprintf(output, print_width, "Type    : Probe request (searching client)");
            break;
        case network_data:
            snprintf(output, print_width, "Type    : Data (no network control traffic)");
            break;
        }
        details_text.push_back(output);
    
        if (dnet->beacon_info.size() > 0) {
            snprintf(output, print_width, "Info    : %s", dnet->beacon_info.c_str());
            details_text.push_back(output);
        }
    
        snprintf(output, print_width, "Channel : %d", dnet->channel);
        details_text.push_back(output);
        snprintf(output, print_width, "WEP     : %s", dnet->wep ? "Yes" : "No");
        details_text.push_back(output);
    
        snprintf(output, print_width, "Beacon  : %d (%f sec)", dnet->beacon,
                 (float) dnet->beacon * 1024 / 1000000);
        details_text.push_back(output);
    
        snprintf(output, print_width, "Packets : %d",
                 dnet->data_packets +
                 dnet->llc_packets +
                 dnet->crypt_packets);
        details_text.push_back(output);
        snprintf(output, print_width, "  Data    : %d", dnet->data_packets);
        details_text.push_back(output);
        snprintf(output, print_width, "  LLC     : %d", dnet->llc_packets);
        details_text.push_back(output);
        snprintf(output, print_width, "  Crypt   : %d", dnet->crypt_packets);
        details_text.push_back(output);
        snprintf(output, print_width, "  Weak    : %d", dnet->interesting_packets);
        details_text.push_back(output);
    
        switch (dnet->atype) {
        case address_none:
            snprintf(output, print_width, "IP Type : None detected");
            break;
        case address_udp:
            snprintf(output, print_width, "IP Type : UDP (%d octets)", dnet->octets);
            break;
        case address_arp:
            snprintf(output, print_width, "IP Type : ARP (%d octets)", dnet->octets);
            break;
        case address_dhcp:
            snprintf(output, print_width, "IP Type : DHCP");
            break;
        case address_group:
            snprintf(output, print_width, "IP Type : Group (aggregate)");
            break;
        }
        details_text.push_back(output);
    
        if (dnet->atype != address_none) {
            snprintf(output, print_width, "IP Range: %d.%d.%d.%d",
                     dnet->range_ip[0], dnet->range_ip[1],
                     dnet->range_ip[2], dnet->range_ip[3]);
            details_text.push_back(output);
    
            if (dnet->atype == address_dhcp) {
                snprintf(output, print_width, "Netmask : %d.%d.%d.%d",
                         dnet->mask[0], dnet->mask[1],
                         dnet->mask[2], dnet->mask[3]);
                details_text.push_back(output);
    
                snprintf(output, print_width, "Gateway : %d.%d.%d.%d",
                        dnet->gate_ip[0], dnet->gate_ip[1],
                        dnet->gate_ip[2], dnet->gate_ip[3]);
                details_text.push_back(output);
            }
    
        }
    
        if (dnet->gps_fixed != -1) {
            snprintf(output, print_width, "Min Loc : Lat %f Lon %f Alt %f Spd %f",
                     dnet->min_lat, dnet->min_lon,
                     dnet->min_alt, dnet->min_spd);
            details_text.push_back(output);
            snprintf(output, print_width, "Max Loc : Lat %f Lon %f Alt %f Spd %f",
                     dnet->max_lat, dnet->max_lon,
                     dnet->max_alt, dnet->max_spd);
            details_text.push_back(output);
        }
    }
    
    for (unsigned int x = 0; x + popup_start < details_text.size() &&
         x < (unsigned int) panel_window(in_panel)->_maxy - 1; x++) {
        mvwaddstr(panel_window(in_panel), 1+x, 3, details_text[x+popup_start].c_str());
        popup_end = x + popup_start;
    }

    popup_length = details_text.size();
}

void PanelFront::PrintKismetDump(PANEL *in_panel) {
    static vector<string> strlist;

    if (clear_dump) {
        client->ClearStrings();
        clear_dump = 0;
    } else if (!popup_paused) {
        strlist = client->FetchStrings();
    }

    unsigned int maxprint = panel_window(in_panel)->_maxy - 1;
    unsigned int maxwidth = panel_window(in_panel)->_maxx - 4;

    if (strlist.size() > maxprint)
        strlist.erase(strlist.begin(), strlist.begin() + (strlist.size() - maxprint));

    char *line = new char[maxwidth];
    for (unsigned int x = 0; x < maxprint && x < strlist.size(); x++) {
        snprintf(line, maxwidth, "%s", strlist[x].c_str());
        mvwaddstr(panel_window(in_panel), 1+x, 3, line);
    }
    free(line);
}

// We're special -- because we capture an entire string of user input, we have to
// do this ourselves.
void PanelFront::PrintKismetGroupname(PANEL *in_panel) {
    char gname[26];
    int print_width = panel_window(in_panel)->_maxx - 5;

    if (print_width > 25)
        print_width = 25;

    werase(panel_window(in_panel));
    box(panel_window(in_panel), '|', '-');
    mvwaddstr(panel_window(in_panel), 0, 2, popup_title.c_str());


    mvwaddstr(panel_window(in_panel), 1, 2, "Group name:");

    // The text field is reversed
    wattron(panel_window(in_panel), WA_REVERSE);

    memset(gname, ' ', print_width);
    gname[print_width] = '\0';

    mvwaddstr(panel_window(in_panel), 2, 2, gname);

    wattroff(panel_window(in_panel), WA_REVERSE);

    if (details_network->name == "")
        snprintf(gname, print_width - 9, "%s", details_network->virtnet.ssid.c_str());
    else
        snprintf(gname, print_width - 9, "%s", details_network->name.c_str());
    mvwaddstr(panel_window(in_panel), 3, 2, "Default: ");
    mvwaddstr(panel_window(in_panel), 3, 11, gname);

    echo();
    nocbreak();
    nodelay(panel_window(in_panel), 0);

    wattron(panel_window(in_panel), WA_REVERSE);
    mvwgetnstr(panel_window(in_panel), 2, 2, gname, print_width-1);
    wattroff(panel_window(in_panel), WA_REVERSE);

    noecho();
    cbreak();

    // Kill the popup ourselves
    hide_panel(in_panel);
    del_panel(in_panel);
    delwin(popupwin);
    popup_panel = NULL;
    cur_panel = network_panel;

    if (strlen(gname) == 0)
        return;

    details_network->name = gname;
    group_name_map[details_network->tag] = gname;
}

int PanelFront::Poll() {
    int ch;

    if ((ch = wgetch(panel_window(cur_panel))) == ERR)
        return 0;

//    char foo[100];

    if (cur_panel == network_panel) {
        switch (ch) {
        case 'q':
        case 'Q':
            return FE_QUIT;
            break;
        case KEY_UP:
            if (sortby != sort_auto) {
                if (selected == 0 && display_start != 0) {
                    display_start--;
                } else if (selected > 0) {
                    selected--;
                }
            }

            /*
            snprintf(foo, 100, "Selected: %d start: %d end: %d",
                     selected, display_start, display_end);
                     WriteStatus(foo);
                     */


            /*
            if (sortby != sort_auto || paused) {
                if (selected == 0 && display_start != 0) {
                    display_start--;
                } else if (selected > 0) {
                    selected--;
                    }
                    */

            break;
        case KEY_DOWN:
        if (sortby != sort_auto) {
            if (display_start + selected < group_vec.size() - 1) {
                if (display_start + selected >= display_end)
                    display_start++;
                else
                    selected++;
            }
        }
        break;
        /*
         case KEY_PPAGE:
            if (sortby != sort_auto) {
                if (display_start > 0)
                    display_start -= (netwin->_maxy - netwin->_begy - 3) / 2;

                if (display_start < 0)
                    display_start = 0;
            }
            break;
        case KEY_NPAGE:
            if (sortby != sort_auto && !paused) {
                if (display_end < client->FetchNumNetworks()) {
                    display_start += (netwin->_maxy - netwin->_begy - 3) / 2;

                    if (display_start > client->FetchNumNetworks() - 1)
                        display_start = client->FetchNumNetworks() - 1;

                }
            }
            break;
            */
        case KEY_RIGHT:
        case '+':
            if (sortby != sort_auto) {
                if (last_displayed[selected]->type == group_bundle)
                    last_displayed[selected]->expanded = 1;
            }
            break;
        case KEY_LEFT:
        case '-':
            if (sortby != sort_auto) {
                if (last_displayed[selected]->type == group_bundle)
                    last_displayed[selected]->expanded = 0;
            }
            break;
        case 'i':
        case 'I':
        case KEY_ENTER:
            if (sortby != sort_auto) {
                details_network = last_displayed[selected];
                SpawnPopup("Network Details", &PanelFront::PrintKismetInfo);
            }
            break;
        case 't':
        case 'T':
            if (sortby != sort_auto) {
                if (last_displayed[selected]->tagged)
                    last_displayed[selected]->tagged = 0;
                else
                    last_displayed[selected]->tagged = 1;
            }
            break;
        case 'n':
        case 'N':
            if (sortby != sort_auto) {
                details_network = last_displayed[selected];
                SpawnPopup("Group Name", &PanelFront::PrintKismetGroupname, 5, 30);
            }
            break;
        case 'g':
        case 'G':
            details_network = GroupTagged();
            if (details_network != NULL)
                SpawnPopup("Group Name", &PanelFront::PrintKismetGroupname, 5, 30);
            break;
        case 'u':
        case 'U':
            DestroyGroup(last_displayed[selected]);
            break;
        case 'h':
        case 'H':
            SpawnPopup("Kismet Help", &PanelFront::PrintKismetHelp, HELP_SIZE);
            break;
        case 'z':
        case 'Z':
            ZoomNetworks();
            break;
        case 's':
        case 'S':
            sub_sort = 1;
            SpawnPopup("Sort Network", &PanelFront::PrintKismetSort, SORT_SIZE);
            break;
        case 'l':
        case 'L':
            SpawnPopup("Wireless Card Power", &PanelFront::PrintKismetPower, 5);
            break;
        case 'd':
        case 'D':
            sub_dump = 1;
            SpawnPopup("Data Strings Dump", &PanelFront::PrintKismetDump);
            break;
        case 'm':
        case 'M':
            if (muted) {
                speech = old_speech;
                sound = old_sound;
                muted = 0;
                WriteStatus("Restoring sound");
            } else if (sound != 0 || speech != 0) {
                old_speech = speech;
                old_sound = sound;
                sound = 0;
                speech = 0;
                muted = 1;
                WriteStatus("Muting sound");
            } else if (sound == 0 && speech == 0) {
                WriteStatus("Sound not enabled.");
            }
            break;
        }
    } else if (cur_panel == popup_panel) {
        if (sub_sort) {
            switch (ch) {
            case 'a':
            case 'A':
                sortby = sort_auto;
                WriteStatus("Autofitting network display");
                break;
            case 'c':
            case 'C':
                sortby = sort_channel;
                WriteStatus("Sorting by channel");
                break;
            case 'f':
                sortby = sort_first;
                WriteStatus("Sorting by time first detected");
                break;
            case 'F':
                sortby = sort_first_dec;
                WriteStatus("Sorting by time first detected (descending)");
                break;
            case 'l':
                sortby = sort_last;
                WriteStatus("Sorting by time most recently active");
                break;
            case 'L':
                sortby = sort_last_dec;
                WriteStatus("Sorting by time most recently active (descending)");
                break;
            case 'b':
                sortby = sort_bssid;
                WriteStatus("Sorting by BSSID");
                break;
            case 'B':
                sortby = sort_bssid_dec;
                WriteStatus("Sorting by BSSID (descending)");
                break;
            case 's':
                sortby = sort_ssid;
                WriteStatus("Sorting by SSID");
                break;
            case 'S':
                sortby = sort_ssid_dec;
                WriteStatus("Sorting by SSID (descending)");
                break;
            case 'w':
            case 'W':
                sortby = sort_wep;
                WriteStatus("Sorting by WEP");
                break;
            case 'p':
                sortby = sort_packets;
                WriteStatus("Sorting by packet counts.");
                break;
            case 'P':
                sortby = sort_packets_dec;
                WriteStatus("Sorting by packet counts (descending)");
                break;
            case 'x':
            case 'X':
                break;
            default:
                beep();
                return 0;
                break;
            }

            sub_sort = 0;

            hide_panel(cur_panel);
            del_panel(cur_panel);
            delwin(popupwin);
            cur_panel = network_panel;

            DrawDisplay();

            return 0;
        } else if (sub_dump) {
            switch(ch) {
            case 'p':
            case 'P':
                // Ignore if we're pending a clear
                if (clear_dump == 1)
                    break;

                if (popup_paused)
                    popup_paused = 0;
                else
                    popup_paused = 1;
                break;

            case 'c':
            case 'C':
                if (!popup_paused)
                    clear_dump = 1;
                break;

            case 'x':
            case 'X':
            case 'q':
            case 'Q':
                hide_panel(cur_panel);
                del_panel(cur_panel);
                delwin(popupwin);
                cur_panel = network_panel;
                sub_dump = 0;
                break;

            default:
                break;
            }

            DrawDisplay();

            return 0;
        }

        switch (ch) {
        case KEY_UP:
            if (popup_start != 0) {
                popup_start--;
            }
            break;
        case KEY_DOWN:
            if (popup_end < popup_length-1 && popup_end != 0) {
                popup_start++;
            }
            break;

        case 'p':
        case 'P':
            if (popup_paused)
                popup_paused = 0;
            else
                popup_paused = 1;
            break;

        case 'x':
        case 'X':
        case 'q':
        case 'Q':
            hide_panel(cur_panel);
            del_panel(cur_panel);
            delwin(popupwin);
            cur_panel = network_panel;
            break;
        }
    }

    DrawDisplay();

    return 0;
}


#endif
